module L = Llvm
module A = Ast
open Sast
module StringMap = Map.Make (String)

let translate_program fdecl_list =
  let context = L.global_context () in

  (* Create the LLVM compilation module into which we will generate code *)
  let the_module = L.create_module context "CTeX" in

  (* Get types from the context *)
  let i32_t = L.i32_type context
  and i8_t = L.i8_type context
  and i1_t = L.i1_type context
  and float_t = L.double_type context in

  (* Return the LLVM type for CTeX type *)
  let ltype_of_typ = function
    | Int -> i32_t
    | Bool -> i1_t
    | Float -> float_t
  in

  (* GNU C BUILTIN FUNCTIONS *)
  let printf_t : L.lltype =
    L.var_arg_function_type i32_t [| L.pointer_type i8_t |]
  in
  let printf_func : L.llvalue =
    L.declare_function "printf" printf_t the_module
  in
  let abs_t = L.function_type i32_t [| i32_t |] in
  let abs_func = L.declare_function "abs" abs_t the_module in
  let fabs_t = L.function_type float_t [| float_t |] in
  let fabs_func = L.declare_function "fabs" fabs_t the_module in
  let floor_t = L.function_type i32_t [| float_t |] in
  let floor_func = L.declare_function "abs" floor_t the_module in
  let ceil_t = L.function_type i32_t [| float_t |] in
  let ceil_func = L.declare_function "abs" ceil_t the_module in
  let pow_t = L.function_type float_t [| float_t; float_t |] in
  let pow_func = L.declare_function "pow" pow_t the_module in
  let lg_t = L.function_type float_t [| float_t |] in
  let lg_func = L.declare_function "log10" lg_t the_module in
  let ln_t = L.function_type float_t [| float_t |] in
  let ln_func = L.declare_function "log" ln_t the_module in
  let sqrt_t = L.function_type float_t [| float_t |] in
  let sqrt_func = L.declare_function "sqrt" sqrt_t the_module in
  let sin_t = L.function_type float_t [| float_t |] in
  let sin_func = L.declare_function "sin" sin_t the_module in
  let cos_t = L.function_type float_t [| float_t |] in
  let cos_func = L.declare_function "cos" cos_t the_module in
  let tan_t = L.function_type float_t [| float_t |] in
  let tan_func = L.declare_function "tan" tan_t the_module in
  let arcsin_t = L.function_type float_t [| float_t |] in
  let arcsin_func = L.declare_function "asin" arcsin_t the_module in
  let arccos_t = L.function_type float_t [| float_t |] in
  let arccos_func = L.declare_function "acos" arccos_t the_module in
  let arctan_t = L.function_type float_t [| float_t |] in
  let arctan_func = L.declare_function "atan" arctan_t the_module in
  let sinh_t = L.function_type float_t [| float_t |] in
  let sinh_func = L.declare_function "sinh" sinh_t the_module in
  let cosh_t = L.function_type float_t [| float_t |] in
  let cosh_func = L.declare_function "cosh" cosh_t the_module in
  let tanh_t = L.function_type float_t [| float_t |] in
  let tanh_func = L.declare_function "tanh" tanh_t the_module in

  (* EXTENDED BUILTIN FUNCTION *)
  let binom_t = L.function_type i32_t [| i32_t; i32_t |] in
  let binom_func = L.declare_function "binom" binom_t the_module in
  let gcd_t = L.function_type i32_t [| i32_t; i32_t |] in
  let gcd_func = L.declare_function "gcd" gcd_t the_module in

  let function_decls : (L.llvalue * L.lltype * sfunc_decl) StringMap.t =
    let function_decl m fdecl =
      let name = fdecl.sfname
      and formal_types =
        Array.of_list
          (if List.length fdecl.sframe == 0 then
           List.map (fun (t, _) -> ltype_of_typ t) fdecl.sformals
          else
            let func_frame_type =
              L.named_struct_type context (fdecl.sfname ^ "_frame")
            in
            L.struct_set_body func_frame_type
              (Array.of_list
                 (List.map (fun (t, _) -> ltype_of_typ t) fdecl.sframe))
              false;
            func_frame_type
            :: List.map (fun (t, _) -> ltype_of_typ t) fdecl.sformals)
      in
      let ftype = L.function_type (ltype_of_typ fdecl.styp) formal_types in
      StringMap.add name
        (L.define_function name ftype the_module, ftype, fdecl)
        m
    in
    List.fold_left function_decl StringMap.empty fdecl_list
  in

  (* MAIN ENTRY BUILDER *)
  let build_func fdecl =
    let the_function, the_types, _ =
      StringMap.find fdecl.sfname function_decls
    in
    let builder = L.builder_at_end context (L.entry_block the_function) in
    (* Construct the function's "locals": formal arguments and locally
       declared variables. Allocate each on the stack, initialize their
       value, if appropriate, and remember their values in the "locals"
       map *)
    let local_vars =
      let add_formal m (t, n) p =
        L.set_value_name n p;
        let local = L.build_alloca (ltype_of_typ t) n builder in
        ignore (L.build_store p local builder);
        StringMap.add n local m
      and add_frame_element m (t, n) p =
        let frame_element_ptr =
          L.build_struct_gep p (StringMap.cardinal m)
            (fdecl.sfname ^ "_framearg_"
            ^ string_of_int (StringMap.cardinal m))
            builder
        in
        let frame_element_p = L.build_alloca (ltype_of_typ t) n builder in
        let frame_element_v = L.build_load frame_element_ptr n builder in
        ignore (L.build_store frame_element_v frame_element_p builder);
        StringMap.add n frame_element_p m
      (* Allocate space for any locally declared variables and add the
       * resulting registers to our map *)
      and add_local m (t, n) =
        let local_var = L.build_alloca (ltype_of_typ t) n builder in
        StringMap.add n local_var m
      in
      if List.length fdecl.sframe == 0 then
        let formals =
          List.fold_left2 add_formal StringMap.empty fdecl.sformals
            (Array.to_list (L.params the_function))
        in
        List.fold_left add_local formals fdecl.slocals
      else
        let frame_arg = List.hd (Array.to_list (L.params the_function)) in
        let frame_type = List.hd (Array.to_list (L.param_types the_types)) in
        let frame =
          L.build_alloca frame_type (fdecl.sfname ^ "_frame_ptr") builder
        in
        ignore (L.build_store frame_arg frame builder);
        let frame_elements =
          List.fold_left
            (fun m (t, n) -> add_frame_element m (t, n) frame)
            StringMap.empty fdecl.sframe
        in
        let formals =
          List.fold_left2 add_formal frame_elements fdecl.sformals
            (List.tl (Array.to_list (L.params the_function)))
        in
        List.fold_left add_local formals fdecl.slocals
    in

    (* Return the value for a variable or formal argument. Check local names
       first, then global names *)
    let lookup lb n =
      try StringMap.find n lb
      with Not_found -> raise (Failure ("Unbound Variable " ^ n))
    in
    let add_terminal builder instr =
      match L.block_terminator (L.insertion_block builder) with
      | Some _ -> ()
      | None -> ignore (instr builder)
    in
    let rec trans_atom builder lb ((t, se) : s_atom) =
      match se with
      | SId s -> L.build_load (lookup lb s) s builder
      | SIntLiteral l -> L.const_int i32_t l
      | SFloatLiteral l -> L.const_float float_t l
      | SParen e -> trans_expr_calc builder lb e
      | SParenthLike (op, (et, e)) -> (
          let e' = trans_expr_calc builder lb (et, e) in
          match op with
          | A.Abs -> (
              match et with
              | Int -> L.build_call abs_func [| e' |] "abs" builder
              | Float -> L.build_call fabs_func [| e' |] "fabs" builder
              | Bool ->
                  raise
                    (Failure "Compile-time Error: Bool type found in abs"))
          | A.Floor -> (
              match et with
              | Int -> e'
              | Float -> L.build_call floor_func [| e' |] "floor" builder
              | Bool ->
                  raise
                    (Failure "Compile-time Error: Bool type found in floor"))
          | A.Ceil -> (
              match et with
              | Int -> e'
              | Float -> L.build_call ceil_func [| e' |] "ceil" builder
              | Bool ->
                  raise
                    (Failure "Compile-time Error: Bool type found in ceil")))
    and trans_expr_pow builder lb ((t, se) : s_expr_pow) =
      match se with
      | SAtom a -> trans_atom builder lb a
      | SPowerFunc ((t1, a1), (t2, a2)) -> (
          let a1' = trans_atom builder lb (t1, a1)
          and a2' = trans_expr_calc builder lb (t2, a2) in
          match (t1, t2) with
          | Int, Int ->
              L.build_call pow_func
                [|
                  L.build_sitofp a1' float_t "pow_tmp_a1" builder;
                  L.build_sitofp a2' float_t "pow_tmp_a2" builder;
                |]
                "pow" builder
          | Int, Float ->
              L.build_call pow_func
                [| L.build_sitofp a1' float_t "pow_tmp_a1" builder; a2' |]
                "pow" builder
          | Float, Int ->
              L.build_call pow_func
                [| a1'; L.build_sitofp a2' float_t "pow_tmp_a2" builder |]
                "pow" builder
          | Float, Float ->
              L.build_call pow_func [| a1'; a2' |] "pow" builder
          | _ -> raise (Failure "Compile-time Error: Bool type found in pow")
          )
    and trans_expr_log builder lb ((t, se) : s_expr_log) =
      match se with
      | SExprPow e -> trans_expr_pow builder lb e
      | SLogLike (op, (et, e)) -> (
          let e' = trans_expr_log builder lb (et, e) in
          let e' =
            match et with
            | Int -> L.build_sitofp e' float_t "log_like_tmp" builder
            | Float -> e'
            | Bool ->
                raise
                  (Failure
                     "Compile-time Error: Bool type found in log like \
                      function")
          in
          match op with
          | A.Lg -> L.build_call lg_func [| e' |] "log10" builder
          | A.Ln -> L.build_call ln_func [| e' |] "log" builder
          | A.Sqrt -> L.build_call sqrt_func [| e' |] "sqrt" builder
          | A.Sin -> L.build_call sin_func [| e' |] "sin" builder
          | A.Cos -> L.build_call cos_func [| e' |] "cos" builder
          | A.Tan -> L.build_call tan_func [| e' |] "tan" builder
          | A.Arcsin -> L.build_call arcsin_func [| e' |] "asin" builder
          | A.Arccos -> L.build_call arccos_func [| e' |] "acos" builder
          | A.Arctan -> L.build_call arctan_func [| e' |] "atan" builder
          | A.Sinh -> L.build_call sinh_func [| e' |] "sinh" builder
          | A.Cosh -> L.build_call cosh_func [| e' |] "cosh" builder
          | A.Tanh -> L.build_call tanh_func [| e' |] "tanh" builder
          | A.Csc ->
              L.build_fdiv
                (L.const_float float_t 1.0)
                (L.build_call sin_func [| e' |] "sin" builder)
                "csc" builder
          | A.Sec ->
              L.build_fdiv
                (L.const_float float_t 1.0)
                (L.build_call cos_func [| e' |] "cos" builder)
                "sec" builder
          | A.Cot ->
              L.build_fdiv
                (L.const_float float_t 1.0)
                (L.build_call tan_func [| e' |] "tan" builder)
                "cot" builder
          | A.Coth ->
              L.build_fdiv
                (L.const_float float_t 1.0)
                (L.build_call tanh_func [| e' |] "tanh" builder)
                "coth" builder)
      | SLog ((t1, a), (t2, e)) ->
          let a' = trans_expr_calc builder lb (t1, a)
          and e' = trans_expr_log builder lb (t2, e) in
          let a' =
            match t1 with
            | Int -> L.build_sitofp a' float_t "log_base_tmp" builder
            | Float -> a'
            | Bool ->
                raise
                  (Failure
                     "Compile-time Error: Bool type found in log function")
          and e' =
            match t2 with
            | Int -> L.build_sitofp e' float_t "log_value_tmp" builder
            | Float -> e'
            | Bool ->
                raise
                  (Failure
                     "Compile-time Error: Bool type found in log function")
          in
          L.build_fdiv
            (L.build_call ln_func [| e' |] "log_value" builder)
            (L.build_call ln_func [| a' |] "log_base" builder)
            "log" builder
    and trans_expr_impl_mult builder lb ((t, se) : s_expr_impl_mult) =
      match se with
      | SExprLog e -> trans_expr_log builder lb e
      | SImplMult ((t1, e1), (t2, e2)) -> (
          let e1' = trans_expr_impl_mult builder lb (t1, e1)
          and e2' = trans_expr_log builder lb (t2, e2) in
          match (t1, t2) with
          | Int, Int -> L.build_mul e1' e2' "impl_tmp" builder
          | Float, Int ->
              L.build_fmul e1'
                (L.build_sitofp e2' float_t "impl_tmp_e2" builder)
                "impl_tmp" builder
          | Int, Float ->
              L.build_fmul
                (L.build_sitofp e1' float_t "impl_tmp_e1" builder)
                e2' "impl_tmp" builder
          | Float, Float -> L.build_fmul e1' e2' "impl_tmp" builder
          | _ ->
              raise
                (Failure
                   "Compile-time Error: Bool type found in implicit multiply")
          )
      | SCall (fname, args) -> (
          match fname with
          | A.FId s ->
              let fdef, ft, fdecl = StringMap.find (A.string_of_ident s) function_decls in
              let trans_arg (t, e) =
                let e' = trans_expr_calc builder lb (t, e) in
                match t with
                | Int -> L.build_sitofp e' float_t "arg_value_tmp" builder
                | Float -> e'
                | Bool ->
                    raise
                      (Failure
                         "Compile-time Error: Bool type found in function \
                          args")
              in
              let llargs =
                if List.length fdecl.sframe != 0 then
                  let frame =
                    L.build_load
                      (lookup lb (A.string_of_ident s ^ "_frame_ptr"))
                      (A.string_of_ident s ^ "_frame") builder
                  in
                  frame :: List.rev (List.map trans_arg (List.rev args))
                else List.rev (List.map trans_arg (List.rev args))
              in
              L.build_call fdef (Array.of_list llargs) (A.string_of_ident s ^ "_result")
                builder
          | A.Gcd ->
              let t1, e1 = List.hd args
              and t2, e2 = List.hd (List.tl (List.rev args)) in
              let e1' = trans_expr_calc builder lb (t1, e1)
              and e2' = trans_expr_calc builder lb (t2, e2) in
              let e1' =
                match t1 with
                | Int -> e1'
                | Float -> L.build_fptoui e1' i32_t "gcd_tmp_e1" builder
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in gcd")
              and e2' =
                match t2 with
                | Int -> e2'
                | Float -> L.build_fptoui e2' i32_t "gcd_tmp_e2" builder
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in gcd")
              in
              L.build_call gcd_func [| e1'; e2' |] "gcd" builder
          | A.Min | A.Max ->
              let rec build_minmax args =
                match args with
                | [ (t, e) ] -> (t, trans_expr_calc builder lb (t, e))
                | (t, e) :: tl ->
                    let e = trans_expr_calc builder lb (t, e)
                    and rt, rv = build_minmax tl in
                    let cond =
                      match (t, rt, fname) with
                      | Int, Int, A.Min ->
                          L.build_icmp L.Icmp.Sle e rv "leq_tmp" builder
                      | Int, Int, A.Max ->
                          L.build_icmp L.Icmp.Sge e rv "geq_tmp" builder
                      | Float, Int, A.Min ->
                          L.build_fcmp L.Fcmp.Ole e
                            (L.build_sitofp rv float_t "leq_tmp_e2" builder)
                            "leq_tmp" builder
                      | Float, Int, A.Max ->
                          L.build_fcmp L.Fcmp.Oge e
                            (L.build_sitofp rv float_t "geq_tmp_e2" builder)
                            "geq_tmp" builder
                      | Int, Float, A.Min ->
                          L.build_fcmp L.Fcmp.Ole
                            (L.build_sitofp e float_t "leq_tmp_e1" builder)
                            rv "leq_tmp" builder
                      | Int, Float, A.Max ->
                          L.build_fcmp L.Fcmp.Oge
                            (L.build_sitofp e float_t "geq_tmp_e1" builder)
                            rv "geq_tmp" builder
                      | Float, Float, A.Min ->
                          L.build_fcmp L.Fcmp.Ole e rv "leq_tmp" builder
                      | Float, Float, A.Max ->
                          L.build_fcmp L.Fcmp.Oge e rv "geq_tmp" builder
                      | _ ->
                          raise
                            (Failure
                               "Compile-time Error: Bool type found in comp")
                    in
                    let typ = if t == Int && rt == Int then Int else Float in
                    (typ, L.build_select cond e rv "minmax_tmp" builder)
                | [] -> failwith "no element"
              in
              snd (build_minmax args))
    and trans_expr_unary builder lb ((t, se) : s_expr_unary) =
      match se with
      | SExprImplMult e -> trans_expr_impl_mult builder lb e
      | SUnop (op, (et, e)) -> (
          let ee = trans_expr_unary builder lb (et, e) in
          match op with
          | A.Uplus -> ee
          | A.Uminus -> (
              match et with
              | Int -> L.build_neg ee "unary_tmp" builder
              | Float -> L.build_fneg ee "unary_tmp" builder
              | Bool ->
                  raise
                    (Failure "Compile-time Error: Bool type found in neg")))
      | SId s -> L.build_load (lookup lb s) s builder
      | SParen e -> trans_expr_calc builder lb e
    and trans_expr_mult builder lb ((t, se) : s_expr_mult) =
      match se with
      | SExprUnary e -> trans_expr_unary builder lb e
      | SMultLike ((t1, e1), op, (t2, e2)) -> (
          let e1' = trans_expr_mult builder lb (t1, e1)
          and e2' = trans_expr_unary builder lb (t2, e2) in
          match op with
          | A.Mult -> (
              match (t1, t2) with
              | Int, Int -> L.build_mul e1' e2' "mul_tmp" builder
              | Float, Int ->
                  L.build_fmul e1'
                    (L.build_sitofp e2' float_t "mul_tmp_e2" builder)
                    "mul_tmp" builder
              | Int, Float ->
                  L.build_fmul
                    (L.build_sitofp e1' float_t "mul_tmp_e1" builder)
                    e2' "mul_tmp" builder
              | Float, Float -> L.build_fmul e1' e2' "mul_tmp" builder
              | _ ->
                  raise
                    (Failure
                       "Compile-time Error: Bool type found in multiply"))
          | A.Div ->
              let e1' =
                match t1 with
                | Int -> L.build_sitofp e1' float_t "div_tmp_e1" builder
                | Float -> e1'
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in div")
              and e2' =
                match t2 with
                | Int -> L.build_sitofp e2' float_t "div_tmp_e2" builder
                | Float -> e2'
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in div")
              in
              L.build_fdiv e1' e2' "div_tmp" builder
          | A.Mod ->
              let e1' =
                match t1 with
                | Int -> e1'
                | Float -> L.build_fptoui e1' i32_t "mod_tmp_e1" builder
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in mod")
              and e2' =
                match t2 with
                | Int -> e2'
                | Float -> L.build_fptoui e2' i32_t "mod_tmp_e2" builder
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in mod")
              in
              L.build_urem e1' e2' "mod_tmp" builder)
    and trans_expr_add builder lb ((t, se) : s_expr_add) =
      match se with
      | SExprMult e -> trans_expr_mult builder lb e
      | SAddLike ((t1, e1), op, (t2, e2)) -> (
          let e1' = trans_expr_add builder lb (t1, e1)
          and e2' = trans_expr_mult builder lb (t2, e2) in
          match op with
          | A.Plus -> (
              match (t1, t2) with
              | Int, Int -> L.build_add e1' e2' "add_tmp" builder
              | Float, Int ->
                  L.build_fadd e1'
                    (L.build_sitofp e2' float_t "add_tmp_e2" builder)
                    "add_tmp" builder
              | Int, Float ->
                  L.build_fadd
                    (L.build_sitofp e1' float_t "add_tmp_e1" builder)
                    e2' "add_tmp" builder
              | Float, Float -> L.build_fadd e1' e2' "add_tmp" builder
              | _ ->
                  raise
                    (Failure "Compile-time Error: Bool type found in add"))
          | A.Minus -> (
              match (t1, t2) with
              | Int, Int -> L.build_sub e1' e2' "sub_tmp" builder
              | Float, Int ->
                  L.build_fsub e1'
                    (L.build_sitofp e2' float_t "sub_tmp_e2" builder)
                    "sub_tmp" builder
              | Int, Float ->
                  L.build_fsub
                    (L.build_sitofp e1' float_t "sub_tmp_e1" builder)
                    e2' "sub_tmp" builder
              | Float, Float -> L.build_fsub e1' e2' "sub_tmp" builder
              | _ ->
                  raise
                    (Failure "Compile-time Error: Bool type found in add")))
    and trans_expr_calc builder lb ((t, se) : s_expr_calc) =
      match se with
      | SExprAdd e -> trans_expr_add builder lb e
      | SCall (fname, args) -> (
          match fname with
          | A.FId s ->
              let fdef, ft, fdecl = StringMap.find (A.string_of_ident s) function_decls in
              let trans_arg (t, e) =
                let e' = trans_expr_calc builder lb (t, e) in
                match t with
                | Int -> L.build_sitofp e' float_t "arg_value_tmp" builder
                | Float -> e'
                | Bool ->
                    raise
                      (Failure
                         "Compile-time Error: Bool type found in function \
                          args")
              in
              let llargs =
                if List.length fdecl.sframe != 0 then
                  let frame =
                    L.build_load
                      (lookup lb (A.string_of_ident s ^ "_frame_ptr"))
                      (A.string_of_ident s ^ "_frame") builder
                  in
                  frame :: List.rev (List.map trans_arg (List.rev args))
                else List.rev (List.map trans_arg (List.rev args))
              in
              L.build_call fdef (Array.of_list llargs) (A.string_of_ident s ^ "_result")
                builder
          | A.Gcd ->
              let t1, e1 = List.hd args
              and t2, e2 = List.hd (List.tl (List.rev args)) in
              let e1' = trans_expr_calc builder lb (t1, e1)
              and e2' = trans_expr_calc builder lb (t2, e2) in
              let e1' =
                match t1 with
                | Int -> e1'
                | Float -> L.build_fptoui e1' i32_t "gcd_tmp_e1" builder
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in gcd")
              and e2' =
                match t2 with
                | Int -> e2'
                | Float -> L.build_fptoui e2' i32_t "gcd_tmp_e2" builder
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in gcd")
              in
              L.build_call gcd_func [| e1'; e2' |] "gcd" builder
          | A.Min | A.Max ->
              let rec build_minmax args =
                match args with
                | [ (t, e) ] -> (t, trans_expr_calc builder lb (t, e))
                | (t, e) :: tl ->
                    let e = trans_expr_calc builder lb (t, e)
                    and rt, rv = build_minmax tl in
                    let cond =
                      match (t, rt, fname) with
                      | Int, Int, A.Min ->
                          L.build_icmp L.Icmp.Sle e rv "leq_tmp" builder
                      | Int, Int, A.Max ->
                          L.build_icmp L.Icmp.Sge e rv "geq_tmp" builder
                      | Float, Int, A.Min ->
                          L.build_fcmp L.Fcmp.Ole e
                            (L.build_sitofp rv float_t "leq_tmp_e2" builder)
                            "leq_tmp" builder
                      | Float, Int, A.Max ->
                          L.build_fcmp L.Fcmp.Oge e
                            (L.build_sitofp rv float_t "geq_tmp_e2" builder)
                            "geq_tmp" builder
                      | Int, Float, A.Min ->
                          L.build_fcmp L.Fcmp.Ole
                            (L.build_sitofp e float_t "leq_tmp_e1" builder)
                            rv "leq_tmp" builder
                      | Int, Float, A.Max ->
                          L.build_fcmp L.Fcmp.Oge
                            (L.build_sitofp e float_t "geq_tmp_e1" builder)
                            rv "geq_tmp" builder
                      | Float, Float, A.Min ->
                          L.build_fcmp L.Fcmp.Ole e rv "leq_tmp" builder
                      | Float, Float, A.Max ->
                          L.build_fcmp L.Fcmp.Oge e rv "geq_tmp" builder
                      | _ ->
                          raise
                            (Failure
                               "Compile-time Error: Bool type found in comp")
                    in
                    let typ = if t == Int && rt == Int then Int else Float in
                    (typ, L.build_select cond e rv "minmax_tmp" builder)
                | [] -> failwith "no element"
              in
              snd (build_minmax args))
      | SFracLike (op, (t1, a1), (t2, a2)) -> (
          let e1' = trans_expr_calc builder lb (t1, a1)
          and e2' = trans_expr_calc builder lb (t2, a2) in
          match op with
          | A.Frac ->
              let e1' =
                match t1 with
                | Int -> L.build_sitofp e1' float_t "frac_tmp_e1" builder
                | Float -> e1'
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in frac")
              and e2' =
                match t2 with
                | Int -> L.build_sitofp e2' float_t "frac_tmp_e2" builder
                | Float -> e2'
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in frac")
              in
              L.build_fdiv e1' e2' "frac_tmp" builder
          | A.Binom ->
              let e1' =
                match t1 with
                | Int -> e1'
                | Float -> L.build_fptoui e1' i32_t "div_tmp_e1" builder
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in mod")
              and e2' =
                match t2 with
                | Int -> e2'
                | Float -> L.build_fptoui e2' i32_t "div_tmp_e2" builder
                | Bool ->
                    raise
                      (Failure "Compile-time Error: Bool type found in mod")
              in
              L.build_call binom_func [| e1'; e2' |] "binom" builder)
(*      | SLargeOp (op, (ti, i), (tstart, estart), (tend, aend), (texp, exp))*)
(*        ->*)
(*          raise (Failure "Unimplmented")*)
    (* let acc_p = L.build_alloca float_t "acc" builder in let acc =
       L.build_store (L.const_float float_t (if op == A.Sum then 0.0 else
       1.0)) acc_p builder in let piv_p = L.build_alloca i32_t "piv" builder
       in let slb = StringMap.add i piv_p lb in let esv = trans_expr_calc
       builder slb (tstart,estart) in let piv = L.build_store esv piv_p
       builder in let pred_bb = L.append_block context "while" the_function
       in ignore(L.build_br pred_bb builder); let body_bb = L.append_block
       context "while_body" the_function in let body_builder =
       L.builder_at_end context body_bb in let ev = trans_expr_calc
       body_builder slb (texp,exp) in let ev' = if texp == Float then ev else
       L.build_sitofp ev float_t "large_op_tmp_ev" body_builder in let acc =
       if op == A.Sum then L.build_fadd acc ev' "new_acc_tmp" body_builder
       else L.build_fmul acc ev' "new_acc_tmp" body_builder in
       ignore(L.build_store acc acc_p body_builder); let piv = L.build_store
       (L.build_add piv (L.const_int i32_t 1) "new_piv_tmp" body_builder)
       piv_p body_builder in ignore(L.build_br pred_bb body_builder); let
       pred_builder = L.builder_at_end context pred_bb in let aev =
       trans_expr_calc pred_builder slb (tend,aend) in let bool_val =
       L.build_icmp L.Icmp.Sle piv aev "leq_tmp" pred_builder in let merge_bb
       = L.append_block context "merge" the_function in
       ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
       L.builder_at_end context merge_bb *)
    and trans_expr_comp builder lb ((t, se) : s_expr_comp) =
      match se with
      | SComp ((t1, e1), op, (t2, e2)) -> (
          let e1' = trans_expr_calc builder lb (t1, e1)
          and e2' = trans_expr_calc builder lb (t2, e2) in
          match (t1, t2) with
          | Int, Int -> (
              match op with
              | A.Lt -> L.build_icmp L.Icmp.Slt e1' e2' "lt_tmp" builder
              | A.Gt -> L.build_icmp L.Icmp.Sgt e1' e2' "gt_tmp" builder
              | A.Leq -> L.build_icmp L.Icmp.Sle e1' e2' "leq_tmp" builder
              | A.Geq -> L.build_icmp L.Icmp.Sge e1' e2' "geq_tmp" builder
              | A.Eq -> L.build_icmp L.Icmp.Eq e1' e2' "eq_tmp" builder
              | A.Neq -> L.build_icmp L.Icmp.Ne e1' e2' "neq_tmp" builder
              | A.Nmid ->
                  L.build_icmp L.Icmp.Ne
                    (L.build_srem e2' e1' "nmid_mod_tmp" builder)
                    (L.const_int i32_t 0) "nmid_tmp" builder
              | A.Mid ->
                  L.build_icmp L.Icmp.Eq
                    (L.build_srem e2' e1' "mid_mod_tmp" builder)
                    (L.const_int i32_t 0) "mid_tmp" builder)
          | Float, Int -> (
              match op with
              | A.Lt ->
                  L.build_fcmp L.Fcmp.Olt e1'
                    (L.build_sitofp e2' float_t "lt_tmp_e2" builder)
                    "lt_tmp" builder
              | A.Gt ->
                  L.build_fcmp L.Fcmp.Ogt e1'
                    (L.build_sitofp e2' float_t "gt_tmp_e2" builder)
                    "gt_tmp" builder
              | A.Leq ->
                  L.build_fcmp L.Fcmp.Ole e1'
                    (L.build_sitofp e2' float_t "leq_tmp_e2" builder)
                    "leq_tmp" builder
              | A.Geq ->
                  L.build_fcmp L.Fcmp.Oge e1'
                    (L.build_sitofp e2' float_t "ged_tmp_e2" builder)
                    "geq_tmp" builder
              | A.Eq ->
                  L.build_fcmp L.Fcmp.Oeq e1'
                    (L.build_sitofp e2' float_t "eq_tmp_e2" builder)
                    "eq_tmp" builder
              | A.Neq ->
                  L.build_fcmp L.Fcmp.One e1'
                    (L.build_sitofp e2' float_t "neq_tmp_e2" builder)
                    "neq_tmp" builder
              | A.Nmid ->
                  L.build_icmp L.Icmp.Ne
                    (L.build_srem
                       (L.build_fptoui e2' i32_t "nmid_mod_tmp_e1" builder)
                       e1' "nmid_mod_tmp" builder)
                    (L.const_int i32_t 0) "nmid_tmp" builder
              | A.Mid ->
                  L.build_icmp L.Icmp.Eq
                    (L.build_srem
                       (L.build_fptoui e2' i32_t "nmid_mod_tmp_e1" builder)
                       e1' "mid_mod_tmp" builder)
                    (L.const_int i32_t 0) "nmid_tmp" builder)
          | Int, Float -> (
              match op with
              | A.Lt ->
                  L.build_fcmp L.Fcmp.Olt
                    (L.build_sitofp e1' float_t "lt_tmp_e1" builder)
                    e2' "lt_tmp" builder
              | A.Gt ->
                  L.build_fcmp L.Fcmp.Ogt
                    (L.build_sitofp e1' float_t "gt_tmp_e1" builder)
                    e2' "gt_tmp" builder
              | A.Leq ->
                  L.build_fcmp L.Fcmp.Ole
                    (L.build_sitofp e1' float_t "leq_tmp_e1" builder)
                    e2' "leq_tmp" builder
              | A.Geq ->
                  L.build_fcmp L.Fcmp.Oge
                    (L.build_sitofp e1' float_t "ged_tmp_e1" builder)
                    e2' "geq_tmp" builder
              | A.Eq ->
                  L.build_fcmp L.Fcmp.Oeq
                    (L.build_sitofp e1' float_t "eq_tmp_e1" builder)
                    e2' "eq_tmp" builder
              | A.Neq ->
                  L.build_fcmp L.Fcmp.One
                    (L.build_sitofp e1' float_t "neq_tmp_e1" builder)
                    e2' "neq_tmp" builder
              | A.Nmid ->
                  L.build_icmp L.Icmp.Ne
                    (L.build_srem e2'
                       (L.build_fptoui e1' i32_t "nmid_mod_tmp_e2" builder)
                       "nmid_mod_tmp" builder)
                    (L.const_int i32_t 0) "nmid_tmp" builder
              | A.Mid ->
                  L.build_icmp L.Icmp.Eq
                    (L.build_srem e2'
                       (L.build_fptoui e1' i32_t "nmid_mod_tmp_e2" builder)
                       "mid_mod_tmp" builder)
                    (L.const_int i32_t 0) "mid_tmp" builder)
          | Float, Float -> (
              match op with
              | A.Lt -> L.build_fcmp L.Fcmp.Olt e1' e2' "lt_tmp" builder
              | A.Gt -> L.build_fcmp L.Fcmp.Ogt e1' e2' "gt_tmp" builder
              | A.Leq -> L.build_fcmp L.Fcmp.Ole e1' e2' "leq_tmp" builder
              | A.Geq -> L.build_fcmp L.Fcmp.Oge e1' e2' "geq_tmp" builder
              | A.Eq -> L.build_fcmp L.Fcmp.Oeq e1' e2' "eq_tmp" builder
              | A.Neq -> L.build_fcmp L.Fcmp.One e1' e2' "neq_tmp" builder
              | A.Nmid ->
                  L.build_icmp L.Icmp.Ne
                    (L.build_srem
                       (L.build_fptoui e2' i32_t "nmid_mod_tmp_e1" builder)
                       (L.build_fptoui e1' i32_t "nmid_mod_tmp_e2" builder)
                       "nmid_mod_tmp" builder)
                    (L.const_int i32_t 0) "nmid_tmp" builder
              | A.Mid ->
                  L.build_icmp L.Icmp.Eq
                    (L.build_srem
                       (L.build_fptoui e2' i32_t "mid_mod_tmp_e1" builder)
                       (L.build_fptoui e1' i32_t "nmid_mod_tmp_e2" builder)
                       "mid_mod_tmp" builder)
                    (L.const_int i32_t 0) "mid_tmp" builder)
          | _ ->
              raise (Failure "Compile-time Error: Bool type found in comp"))
    and trans_expr_logic_atom builder lb ((t, se) : s_expr_logic_atom) =
      match se with
      | SExprComp e -> trans_expr_comp builder lb e
      | SParenLogic e -> trans_expr_logic builder lb e
      | SNot (t, e) ->
          let e' = trans_expr_logic_atom builder lb (t, e) in
          L.build_not e' "not_tmp" builder
    and trans_expr_logic builder lb ((t, se) : s_expr_logic) =
      match se with
      | SLogicAtom e -> trans_expr_logic_atom builder lb e
      | SLogic ((t1, e1), op, (t2, e2)) ->
          let e1' = trans_expr_logic builder lb (t1, e1)
          and e2' = trans_expr_logic_atom builder lb (t2, e2) in
          (match op with A.And -> L.build_and | A.Or -> L.build_or)
            e1' e2' "logic_tmp" builder
    in
    let rec trans_stmt (builder : L.llbuilder) (lb : L.llvalue StringMap.t)
        (oss : sstmt) =
      match oss with
      | Some (t, ss) -> (
          match ss with
          | SExpr e -> ((builder, lb), trans_expr_calc builder lb e)
          | SAssign (s, e) ->
              let e' = trans_expr_calc builder lb e in
              ignore (L.build_store e' (lookup lb s) builder);
              ((builder, lb), e')
          | SPrint e ->
              let e' = trans_expr_calc builder lb e in
              let fmt =
                match t with
                | Int | Bool ->
                    L.build_global_stringptr "%d\n" "fmt_int" builder
                | Float ->
                    L.build_global_stringptr "%g\n" "fmt_float" builder
              in
              ( (builder, lb),
                L.build_call printf_func [| fmt; e' |] "printf" builder )
          | SFuncDef sfd -> (
              if List.length sfd.sframe != 0 then (
                let _, ft, _ = StringMap.find sfd.sfname function_decls in
                let frame_type =
                  List.hd (Array.to_list (L.param_types ft))
                in
                let frame =
                  L.build_alloca frame_type
                    (sfd.sfname ^ "_frame_ptr")
                    builder
                in
                let store_local index p =
                  let ele_ptr =
                    L.build_struct_gep frame index
                      (sfd.sfname ^ "_framearg_" ^ string_of_int index)
                      builder
                  in
                  let pv =
                    L.build_load p (sfd.sfname ^ "_frameargv") builder
                  in
                  ignore (L.build_store pv ele_ptr builder)
                in
                ignore
                  (List.iteri
                     (fun i (t, e) ->
                       let p = lookup lb e in
                       store_local i p)
                     sfd.sframe);
                let slb =
                  StringMap.add (sfd.sfname ^ "_frame_ptr") frame lb
                in
                ((builder, slb), frame))
              else
                ( (builder, lb),
                  match sfd.styp with
                  | Int | Bool -> L.const_int i32_t 0
                  | Float -> L.const_float_of_string float_t "0.0" ))
          | SCase suites ->
              let result_p =
                match t with
                | Int -> L.build_alloca i32_t "case_result" builder
                | Float -> L.build_alloca float_t "case_result" builder
                | Bool -> raise (Failure "Unsupported type in case")
              in
              let rec build_suties b suites =
                match suites with
                | [ (st, ssuit) ] -> (
                    match ssuit with
                    | SCaseSutie (ss, pred) ->
                        let bool_val = (trans_expr_logic b lb) pred in
                        let merge_bb =
                          L.append_block context "merge_tail" the_function
                        in
                        let merge_b = L.builder_at_end context merge_bb in
                        let then_bb =
                          L.append_block context "then_tail" the_function
                        in
                        let (then_b, lb), then_rll =
                          trans_stmt (L.builder_at_end context then_bb) lb ss
                        in
                        let then_rll =
                            (match (st, t) with
                            | Int, Int | Float, Float -> then_rll
                            | Int, Float ->
                                L.build_sitofp then_rll float_t
                                  "cond_ret_tmp" then_b
                            | Float, Int ->
                                L.build_fptosi then_rll float_t
                                  "cond_ret_tmp" then_b
                            | Bool, _ | _, Bool ->
                                raise (Failure "Unable to return bool"))
                        in

                        ignore (L.build_store then_rll result_p then_b);
                        ignore (L.build_br merge_bb then_b);
                        let else_bb =
                          L.append_block context "else_tail" the_function
                        in
                        ignore
                          (L.build_br merge_bb
                             (L.builder_at_end context else_bb));
                        ignore (L.build_cond_br bool_val then_bb else_bb b);
                        (merge_b, merge_bb))
                | (st, ssuit) :: tl -> (
                    match ssuit with
                    | SCaseSutie (ss, pred) ->
                        let bool_val = (trans_expr_logic b lb) pred in
                        let merge_bb =
                          L.append_block context "merge" the_function
                        in
                        let merge_b = L.builder_at_end context merge_bb in
                        let then_bb =
                          L.append_block context "then" the_function
                        in
                        let (then_b, lb), then_rll =
                          trans_stmt (L.builder_at_end context then_bb) lb ss
                        in
                        let then_rll =
                          (match (st, t) with
                          | Int, Int | Float, Float -> then_rll
                          | Int, Float ->
                              L.build_sitofp then_rll float_t
                                "cond_ret_tmp" then_b
                          | Float, Int ->
                              L.build_fptosi then_rll float_t
                                "cond_ret_tmp" then_b
                          | Bool, _ | _, Bool ->
                              raise (Failure "Unable to return bool"))
                        in
                        ignore (L.build_store then_rll result_p then_b);
                        ignore (L.build_br merge_bb then_b);

                        let else_bb =
                          L.append_block context "else" the_function
                        in
                        let else_b, else_ibb =
                          build_suties (L.builder_at_end context else_bb) tl
                        in
                        ignore (L.build_br merge_bb else_b);
                        ignore (L.build_cond_br bool_val then_bb else_bb b);
                        (merge_b, merge_bb))
                | [] -> failwith "no element"
              in
              let b, bb = build_suties builder suites in
              let b = L.builder_at_end context bb in
              ((b, lb), L.build_load result_p "cond_result" b)
          | SStmtClosure scss ->
              let (b, lb), ll =
                List.fold_left_map
                  (fun (b, lb) s -> trans_stmt b lb s)
                  (builder, lb) scss
              in
              ((b, lb), List.hd (List.rev ll)))
      | None -> raise (Failure "Empty body found in function")
    in

    (* print_string(fdecl.sfname ^ " : " ^ string_of_int (StringMap.cardinal
       local_vars) ^ " " ^ string_of_int (List.length fdecl.sformals)^ " " ^
       string_of_int (List.length fdecl.slocals)); *)
    let (b, lb), rv = trans_stmt builder local_vars fdecl.sbody in
    if compare fdecl.sfname "main" = 0 then
      ignore (L.build_ret (L.const_int i32_t 0) b)
    else
      let t =
        match fdecl.sbody with
        | Some (t, ss) -> t
        | None -> raise (Failure "Empty body found in function")
      in
      add_terminal b
        (match fdecl.styp with
        | Float -> (
            match t with
            | Int -> L.build_ret (L.build_sitofp rv float_t "ret_tmp" b)
            | Float -> L.build_ret rv
            | Bool -> raise (Failure "Unable to return bool"))
        | Int -> (
            match t with
            | Int -> L.build_ret rv
            | Float -> L.build_ret (L.build_fptosi rv i32_t "ret_tmp" b)
            | Bool -> raise (Failure "Unable to return bool"))
        | Bool -> raise (Failure "Unable to return bool"))
  in
  List.iter build_func fdecl_list;
  the_module